//+------------------------------------------------------------------+
//|                                           Schaff Trend Cycle.mq4 |
//|                                                           mladen |
//+------------------------------------------------------------------+
#property copyright "mladen"
#property link      "mladenfx@gmail.com"

#property indicator_separate_window
#property indicator_buffers 3
#property indicator_color1  clrDodgerBlue
#property indicator_color2  clrSandyBrown
#property indicator_color3  clrSandyBrown
#property indicator_width1  2
#property indicator_width2  2
#property indicator_width3  2
#property strict

//
//
//
//
//

extern ENUM_TIMEFRAMES TimeFrame      = PERIOD_CURRENT; // Time frame
extern int            STCPeriod       = 10;          // Schaff period
extern ENUM_APPLIED_PRICE Price       = PRICE_CLOSE; // Price to use
extern int            FastMAPeriod    = 23;          // Fast macd period
extern int            SlowMAPeriod    = 50;          // Slow macd period
extern ENUM_MA_METHOD MacdMethod      = MODE_EMA;    // MA method used for macd calculation
extern double         SignalPeriod    = 3;           // Signal period
extern ENUM_MA_METHOD SignalMethod    = MODE_EMA;    // MA method used for signal calculation
extern bool           alertsOn        = false;       // Turn alerts on
extern bool           alertsOnCurrent = true;        // Alerts on still opened bar
extern bool           alertsMessage   = true;        // Alerts should show popup message
extern bool           alertsSound     = false;       // Alerts should play a sound
extern bool           alertsEmail     = false;       // Alerts should send email
extern bool           alertsPushNotif = false;       // Alerts should send notification
extern bool           Interpolate     = true;        // Interpolate in multi time frame mode?

//
//
//
//
//

double stcBuffer[];
double stcBufferUA[];
double stcBufferUB[];
double macdBuffer[];
double fastKBuffer[];
double fastDBuffer[];
double fastKKBuffer[];
double slope[];
string indicatorFileName;
bool   returnBars;


//------------------------------------------------------------------
//
//------------------------------------------------------------------
//
//
//
//
//

int init()
{
   IndicatorBuffers(8);
      SetIndexBuffer(0,stcBuffer);
      SetIndexBuffer(1,stcBufferUA);
      SetIndexBuffer(2,stcBufferUB);
      SetIndexBuffer(3,macdBuffer);
      SetIndexBuffer(4,fastKBuffer);
      SetIndexBuffer(5,fastDBuffer);
      SetIndexBuffer(6,fastKKBuffer);
      SetIndexBuffer(7,slope);
         indicatorFileName = WindowExpertName();
         returnBars        = TimeFrame==-99;
         TimeFrame         = MathMax(TimeFrame,_Period);
   IndicatorShortName(timeFrameToString(TimeFrame)+" Schaff Trend Cycle ("+(string)STCPeriod+","+(string)FastMAPeriod+","+(string)SlowMAPeriod+")");
   return(0);
}

int deinit()
{
   return(0);
}

//------------------------------------------------------------------
//
//------------------------------------------------------------------
//
//
//
//
//

int start()
{
   int counted_bars=IndicatorCounted();
      if(counted_bars < 0) return(-1);
      if(counted_bars>0) counted_bars--;
         int limit = MathMin(Bars-counted_bars,Bars-1);
         if (returnBars) { stcBuffer[0] = limit+1; return(0); }
         if (TimeFrame != _Period)
         {
            limit = (int)MathMax(limit,MathMin(Bars-1,iCustom(NULL,TimeFrame,indicatorFileName,-99,0,0)*TimeFrame/_Period));
            if (slope[limit] == 1) CleanPoint(limit,stcBufferUA,stcBufferUB);
            for (int i=limit; i>=0; i--)
            {
               int y = iBarShift(NULL,TimeFrame,Time[i]);
                  stcBuffer[i] = iCustom(NULL,TimeFrame,indicatorFileName,PERIOD_CURRENT,STCPeriod,Price,FastMAPeriod,SlowMAPeriod,MacdMethod,SignalPeriod,SignalMethod,alertsOn,alertsOnCurrent,alertsMessage,alertsSound,alertsEmail,alertsPushNotif,0,y);
                  slope[i]     = iCustom(NULL,TimeFrame,indicatorFileName,PERIOD_CURRENT,STCPeriod,Price,FastMAPeriod,SlowMAPeriod,MacdMethod,SignalPeriod,SignalMethod,alertsOn,alertsOnCurrent,alertsMessage,alertsSound,alertsEmail,alertsPushNotif,7,y);
                  stcBufferUA[i] = EMPTY_VALUE;
                  stcBufferUB[i] = EMPTY_VALUE;

                  if (!Interpolate ||  (i>0 && y==iBarShift(NULL,TimeFrame,Time[i-1]))) continue;

                     //
                     //
                     //
                     //
                     //
      
                     int n,k; datetime time = iTime(NULL,TimeFrame,y);
                        for(n = 1; i+n < Bars && Time[i+n] >= time; n++) continue;	
                        for(k = 1; k<n && (i+k)<Bars && (i+n)<Bars; k++)
                           stcBuffer[i+k] = stcBuffer[i] + (stcBuffer[i+n] - stcBuffer[i])*k/n;
            }
            for (int i=limit;i>=0;i--) if (slope[i]==-1) PlotPoint(i,stcBufferUA,stcBufferUB,stcBuffer);
            return(0);
         }

   //
   //
   //
   //
   //
   
   if (slope[limit] == 1) CleanPoint(limit,stcBufferUA,stcBufferUB);
   double alpha = 2.0/(1.0+SignalPeriod);
   for(int i = limit; i >= 0; i--)
   {
      double price = iMA(NULL,0,1,0,MODE_SMA,Price,i);
      macdBuffer[i] = iCustomMa(MacdMethod,price,FastMAPeriod,i,0)-iCustomMa(MacdMethod,price,SlowMAPeriod,i,1);
                      if (i>=Bars-1) continue;

      //
      //
      //
      //
      //
      
      double lowMacd  = minValue(macdBuffer,i);
      double highMacd = maxValue(macdBuffer,i)-lowMacd;
         if (highMacd > 0)
               fastKBuffer[i] = 100*((macdBuffer[i]-lowMacd)/highMacd);
         else  fastKBuffer[i] = fastKBuffer[i+1];
               fastDBuffer[i] = iCustomMa(SignalMethod,fastKBuffer[i],SignalPeriod,i,2);
               
      //
      //
      //
      //
      //
                     
      double lowStoch  = minValue(fastDBuffer,i);
      double highStoch = maxValue(fastDBuffer,i)-lowStoch;
         if (highStoch > 0)
               fastKKBuffer[i] = 100*((fastDBuffer[i]-lowStoch)/highStoch);
         else  fastKKBuffer[i] = fastKKBuffer[i+1];
               stcBuffer[i]    = iCustomMa(SignalMethod,fastKKBuffer[i],SignalPeriod,i,3);
      
         //
         //
         //
         //
         //
      
         stcBufferUA[i] = EMPTY_VALUE;
         stcBufferUB[i] = EMPTY_VALUE;
         slope[i]       = slope[i+1];
            if (stcBuffer[i] > stcBuffer[i+1]) slope[i] =  1;
            if (stcBuffer[i] < stcBuffer[i+1]) slope[i] = -1;
            if (slope[i]==-1) PlotPoint(i,stcBufferUA,stcBufferUB,stcBuffer);
   }   

   //
   //
   //
   //
   //
   
   if (alertsOn)
   {
      int whichBar = 1; if (alertsOnCurrent) whichBar = 0;
      if (slope[whichBar] != slope[whichBar+1])
      if (slope[whichBar] == 1)
            doAlert("up");
      else  doAlert("down");       
   }   
   return(0);
}



//------------------------------------------------------------------
//                                                                  
//------------------------------------------------------------------
//
//
//
//
//

void doAlert(string doWhat)
{
   static string   previousAlert="nothing";
   static datetime previousTime;
   string message;
   
      if (previousAlert != doWhat || previousTime != Time[0]) {
          previousAlert  = doWhat;
          previousTime   = Time[0];

          //
          //
          //
          //
          //

          message =  timeFrameToString(_Period)+" "+Symbol()+" at "+TimeToStr(TimeLocal(),TIME_SECONDS)+" Schaff trend cycle slope changed trend to "+doWhat;
             if (alertsMessage)   Alert(message);
             if (alertsPushNotif) SendNotification(message);
             if (alertsEmail)     SendMail(StringConcatenate(Symbol()," Schaff trend cycle experiment"),message);
             if (alertsSound)     PlaySound("alert2.wav");
      }
}

//-------------------------------------------------------------------
//                                                                  
//-------------------------------------------------------------------
//
//
//
//
//

double iCustomMa(int maMode, double price, double period, int i, int instanceNo=0)
{
   switch(maMode)
   {
      case MODE_SMA:  return(iSma (price,(int)period,Bars-i-1,instanceNo));
      case MODE_EMA:  return(iEma (price,     period,Bars-i-1,instanceNo));
      case MODE_SMMA: return(iSmma(price,     period,Bars-i-1,instanceNo));
      case MODE_LWMA: return(iLwma(price,     period,Bars-i-1,instanceNo));
   }               
   return(0);
}

//
//
//
//
//
   
#define _maInstances 4
#define _maWorkBufferx1 1*_maInstances
#define _maWorkBufferx2 2*_maInstances
#define _maWorkBufferx3 3*_maInstances
#define _maWorkBufferx4 4*_maInstances
#define _maWorkBufferx5 5*_maInstances

double workSma[][_maWorkBufferx2];
double iSma(double price, int period, int r, int instanceNo=0)
{
   if (period<=1) return(price);
   if (ArrayRange(workSma,0)!= Bars) ArrayResize(workSma,Bars); instanceNo *= 2; int k;

   //
   //
   //
   //
   //
      
   workSma[r][instanceNo+0] = price;
   workSma[r][instanceNo+1] = price; for(k=1; k<period && (r-k)>=0; k++) workSma[r][instanceNo+1] += workSma[r-k][instanceNo+0];  
   workSma[r][instanceNo+1] /= 1.0*k;
   return(workSma[r][instanceNo+1]);
}

//
//
//
//
//

double workEma[][_maWorkBufferx1];
double iEma(double price, double period, int r, int instanceNo=0)
{
   if (period<=1) return(price);
   if (ArrayRange(workEma,0)!= Bars) ArrayResize(workEma,Bars);

   //
   //
   //
   //
   //
      
   workEma[r][instanceNo] = price;
   double alpha = 2.0 / (1.0+period);
   if (r>0)
          workEma[r][instanceNo] = workEma[r-1][instanceNo]+alpha*(price-workEma[r-1][instanceNo]);
   return(workEma[r][instanceNo]);
}

//
//
//
//
//

double workSmma[][_maWorkBufferx1];
double iSmma(double price, double period, int r, int instanceNo=0)
{
   if (period<=1) return(price);
   if (ArrayRange(workSmma,0)!= Bars) ArrayResize(workSmma,Bars);

   //
   //
   //
   //
   //

   if (r<period)
         workSmma[r][instanceNo] = price;
   else  workSmma[r][instanceNo] = workSmma[r-1][instanceNo]+(price-workSmma[r-1][instanceNo])/period;
   return(workSmma[r][instanceNo]);
}

//
//
//
//
//

double workLwma[][_maWorkBufferx1];
double iLwma(double price, double period, int r, int instanceNo=0)
{
   if (period<=1) return(price);
   if (ArrayRange(workLwma,0)!= Bars) ArrayResize(workLwma,Bars);
   
   //
   //
   //
   //
   //
   
   workLwma[r][instanceNo] = price;
      double sumw = period;
      double sum  = period*price;

      for(int k=1; k<period && (r-k)>=0; k++)
      {
         double weight = period-k;
                sumw  += weight;
                sum   += weight*workLwma[r-k][instanceNo];  
      }             
      return(sum/sumw);
}

//-------------------------------------------------------------------
//                                                                  
//-------------------------------------------------------------------
//
//
//
//
//

double minValue(double& array[],int shift)
{
   double minValue = array[shift];
            for (int i=1; i<STCPeriod && shift+i<Bars; i++) minValue = MathMin(minValue,array[shift+i]);
   return(minValue);
}
double maxValue(double& array[],int shift)
{
   double maxValue = array[shift];
            for (int i=1; i<STCPeriod && shift+i<Bars; i++) maxValue = MathMax(maxValue,array[shift+i]);
   return(maxValue);
}

//-------------------------------------------------------------------
//                                                                  
//-------------------------------------------------------------------
//
//
//
//
//

void CleanPoint(int i,double& first[],double& second[])
{
   if (i>=Bars-3) return;
   if ((second[i]  != EMPTY_VALUE) && (second[i+1] != EMPTY_VALUE))
        second[i+1] = EMPTY_VALUE;
   else
      if ((first[i] != EMPTY_VALUE) && (first[i+1] != EMPTY_VALUE) && (first[i+2] == EMPTY_VALUE))
          first[i+1] = EMPTY_VALUE;
}

void PlotPoint(int i,double& first[],double& second[],double& from[])
{
   if (i>=Bars-2) return;
   if (first[i+1] == EMPTY_VALUE)
      if (first[i+2] == EMPTY_VALUE) 
            { first[i]  = from[i];  first[i+1]  = from[i+1]; second[i] = EMPTY_VALUE; }
      else  { second[i] =  from[i]; second[i+1] = from[i+1]; first[i]  = EMPTY_VALUE; }
   else     { first[i]  = from[i];                           second[i] = EMPTY_VALUE; }
}

//-------------------------------------------------------------------
//
//-------------------------------------------------------------------
//
//
//
//
//

string sTfTable[] = {"M1","M5","M15","M30","H1","H4","D1","W1","MN"};
int    iTfTable[] = {1,5,15,30,60,240,1440,10080,43200};

string timeFrameToString(int tf)
{
   for (int i=ArraySize(iTfTable)-1; i>=0; i--) 
         if (tf==iTfTable[i]) return(sTfTable[i]);
                              return("");
}